# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)
#
# PySide abstraction layer, partly contributed by J-Fox

import os, platform
from .Logging import dprint

try:
	from PySide6 import QtCore
	from PySide6 import QtWidgets
	from PySide6 import QtGui
	from PySide6.QtWidgets import QApplication
	psversion = 6

except ModuleNotFoundError:
	try:
		import PySide2 # pyright: ignore[reportMissingImports]
		import PIL
		# PySide2>=5.14 is required
		major, minor, *_ = PySide2.__version_info__
		if (major, minor) < (5, 14):
			raise ImportError('Requires PySide2>=5.14', name=PySide2.__package__, path=PySide2.__path__)
		# Pillow<10.0.0 is required
		major, minor = map(int, PIL.__version__.split('.')[:2])
		if (major, minor) >= (10, 0):
			raise ImportError('Requires Pillow<10.0.0 if using PySide2', name=PIL.__package__, path=PIL.__path__)

		dprint("PySide6 cannot be loaded. Using PySide2 code path.")
		from PySide2 import QtCore # pyright: ignore[reportMissingImports]
		from PySide2 import QtWidgets # pyright: ignore[reportMissingImports]
		from PySide2 import QtGui # pyright: ignore[reportMissingImports]
		from PySide2.QtWidgets import QApplication # pyright: ignore[reportMissingImports]
		psversion = 2

		os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
		os.environ["QT_ENABLE_HIGHDPI_SCALING"] = "1"
		QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
		QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

	except ImportError as err2:
		dprint("GUI cannot be loaded.")
		raise err2

	except ModuleNotFoundError as err3:
		dprint("GUI cannot be loaded.")
		raise err3

# exec
try:
	if not callable(getattr(QApplication, "exec", None)):
		QApplication.exec = QApplication.exec_
	if not callable(getattr(QApplication, "exec_", None)):
		QApplication.exec_ = QApplication.exec
except (AttributeError, TypeError):
	pass

# QDesktopWidget
class QDesktopWidget(object):
	def screenGeometry(self, widget):
		if psversion == 2:
			return QtWidgets.QDesktopWidget().screenGeometry()
		else:
			return widget.screen().geometry()

# QActionGroup
if psversion == 2:
	QActionGroup = QtWidgets.QActionGroup
else:
	QActionGroup = QtGui.QActionGroup

# Taskbar Progress
if platform.system() == "Windows":
	if psversion == 2:
		try:
			from PySide2 import QtWinExtras # pyright: ignore[reportMissingImports]
		except ImportError:
			pass
	else:
		import ctypes, types
		from ctypes import POINTER, WINFUNCTYPE, byref, c_int, c_long, c_ulonglong, c_ubyte, c_void_p, c_wchar_p, wintypes

		class _QWinTaskbarProgress:
			def __init__(self):
				self._tb = None
				self._hwnd = None
				self._min, self._max, self._value = 0, 100, 0
				self._visible, self._paused = False, False

			def _bind(self, hwnd):
				self._hwnd = wintypes.HWND(int(hwnd))
				self._tb = c_void_p()
				ole = ctypes.windll.ole32
				ole.CoInitialize(None)
				clsid, iid = (c_ubyte * 16)(), (c_ubyte * 16)()
				ole.CLSIDFromString(c_wchar_p("{56FDF344-FD6D-11D0-958A-006097C9A090}"), clsid)
				ole.CLSIDFromString(c_wchar_p("{EA1AFB91-9E28-4B86-90E9-9E9F8A5EEFAF}"), iid)
				ole.CoCreateInstance(clsid, None, 1, iid, byref(self._tb))
				self._call(3, WINFUNCTYPE(c_long, c_void_p))  # HrInit

			def _call(self, idx, proto, *args):
				return proto(ctypes.cast(self._tb, POINTER(POINTER(c_void_p)))[0][idx])(self._tb, *args)

			def _apply(self):
				if self._tb is None:
					return
				if not self._visible:
					state = 0  # TBPF_NOPROGRESS
				elif self._max - self._min <= 0:
					state = 1  # TBPF_INDETERMINATE
				else:
					state = 8 if self._paused else 2  # TBPF_PAUSED / TBPF_NORMAL
				self._call(10, WINFUNCTYPE(c_long, c_void_p, wintypes.HWND, c_int), self._hwnd, state)
				if state in (2, 8):
					span = self._max - self._min
					self._call(9, WINFUNCTYPE(c_long, c_void_p, wintypes.HWND, c_ulonglong, c_ulonglong), self._hwnd, c_ulonglong(max(0, min(span, self._value - self._min))), c_ulonglong(span))

			def setRange(self, minimum, maximum):
				self._min, self._max = int(minimum), int(maximum); self._apply()

			def setValue(self, value):
				self._value = int(value); self._apply()

			def setVisible(self, visible):
				self._visible = bool(visible); self._apply()

			def setPaused(self, paused):
				self._paused = bool(paused); self._apply()

		class _QWinTaskbarButton:
			def __init__(self):
				self._progress = _QWinTaskbarProgress()
			def progress(self):
				return self._progress
			def setWindow(self, window):
				self._progress._bind(int(window.winId()))

		class _QtWin:
			@staticmethod
			def setCurrentProcessExplicitAppUserModelID(app_id):
				ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(c_wchar_p(app_id))

		QtWinExtras = types.ModuleType("QtWinExtras")
		QtWinExtras.QWinTaskbarButton = _QWinTaskbarButton
		QtWinExtras.QtWin = _QtWin
elif platform.system() == "Linux":
	import types

	try:
		if psversion == 2:
			from PySide2 import QtDBus # pyright: ignore[reportMissingImports]
		else:
			from PySide6 import QtDBus
	except ImportError:
		pass
	else:
		class _QWinTaskbarProgress:
			_SIGNAL_PATH = "/com/canonical/unity/launcherentry/1"
			_SIGNAL_INTERFACE = "com.canonical.Unity.LauncherEntry"
			_SIGNAL_NAME = "Update"

			def __init__(self):
				self._min, self._max, self._value = 0, 100, 0
				self._visible, self._paused = False, False
				app_name = os.environ.get("FLASHGBX_DESKTOP_FILE", "").strip()
				app = QtWidgets.QApplication.instance()
				if app_name == "":
					app_name = app.applicationName().strip() if app is not None else ""
				if app_name == "":
					desktop_file_name = getattr(QtGui.QGuiApplication, "desktopFileName", None)
					if callable(desktop_file_name):
						try:
							app_name = str(desktop_file_name()).strip()
						except Exception:
							app_name = ""
				if app_name == "":
					app_name = "flashgbx"
				desktop_file = os.path.basename(app_name)
				if not desktop_file.endswith(".desktop"):
					desktop_file += ".desktop"
				self._app_uri = "application://{:s}".format(desktop_file) if desktop_file else ""
				self._last_payload = None
				self._bus = QtDBus.QDBusConnection.sessionBus()
				self._available = bool(self._app_uri) and bool(self._bus.isConnected())
				if not self._bus.isConnected():
					dprint("Unity Launcher progress disabled: no DBus session bus.")

			def _apply(self):
				if not self._available:
					return
				span = self._max - self._min
				progress_visible = bool(self._visible and span > 0)
				if progress_visible:
					progress = float(self._value - self._min) / float(span)
					progress = max(0.0, min(1.0, progress))
				else:
					progress = 0.0
				state = (progress_visible, float(progress), bool(self._paused and progress_visible))
				if state == self._last_payload:
					return
				self._last_payload = state
				payload = {
					"progress-visible": bool(state[0]),
					"progress": float(state[1]),
				}
				sent = False
				try:
					message = QtDBus.QDBusMessage.createSignal(self._SIGNAL_PATH, self._SIGNAL_INTERFACE, self._SIGNAL_NAME)
					message.setArguments([self._app_uri, payload])
					if self._bus.send(message):
						sent = True
					if not sent:
						dprint("Unity Launcher progress disabled: launcher API not available in this desktop environment.")
				except Exception as err:
					dprint("Unity Launcher progress update failed: {:s}".format(str(err)))

			def setRange(self, minimum, maximum):
				self._min, self._max = int(minimum), int(maximum)
				self._apply()

			def setValue(self, value):
				self._value = int(value)
				self._apply()

			def setVisible(self, visible):
				self._visible = bool(visible)
				self._apply()

			def setPaused(self, paused):
				self._paused = bool(paused)
				self._apply()

		class _QWinTaskbarButton:
			def __init__(self):
				self._progress = _QWinTaskbarProgress()

			def progress(self):
				return self._progress

			def setWindow(self, window):
				pass

		class _QtWin:
			@staticmethod
			def setCurrentProcessExplicitAppUserModelID(app_id):
				pass

		QtWinExtras = types.ModuleType("QtWinExtras")
		QtWinExtras.QWinTaskbarButton = _QWinTaskbarButton
		QtWinExtras.QtWin = _QtWin


__all__ = ['QtCore', 'QtWidgets', 'QtGui', 'QApplication', 'QDesktopWidget', 'QActionGroup']

def GetQtVersion():
	return tuple(map(int, QtCore.qVersion().split(".")))

def IsDarkMode():
	if GetQtVersion() < (6, 5, 0):
		return False
	try:
		scheme = QtGui.QGuiApplication.styleHints().colorScheme()
		if scheme == QtCore.Qt.ColorScheme.Dark:
			return True
	except Exception:
		pass
	return False

def bitmap2pixmap(data, scale_factor=4):
	try:
		from PIL.ImageQt import ImageQt
		from PIL import Image
		data_converted = data.convert("RGBA")
		pixmap = QtGui.QPixmap.fromImage(ImageQt(data_converted.resize(
			(data_converted.width * scale_factor, data_converted.height * scale_factor),
			Image.NEAREST,
		)))
		pixmap.setDevicePixelRatio(scale_factor)
		return pixmap
	except Exception as e:
		dprint("Couldn’t convert bitmap to pixmap. Error: {error}", error=str(e))
		return False



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

import sys, os, io, ast, struct, array, locale, gettext, platform, re, subprocess, glob

OS_LANGUAGE = "en"
CONFIGURED_LANGUAGE = None
TRANSLATION_AUTHOR = None

# ISO 639-1 language codes: code -> (English name, native name)
LANGUAGES = {"aa": ("Afar", "Qafar af"), "ab": ("Abkhazian", "Аԥсшәа"), "ae": ("Avestan", "Avesta"), "af": ("Afrikaans", "Afrikaans"), "ak": ("Akan", "Akan"), "am": ("Amharic", "አማርኛ"), "an": ("Aragonese", "Aragonés"), "ar": ("Arabic", "العربية"), "as": ("Assamese", "অসমীয়া"), "av": ("Avaric", "авар мацӀ"), "ay": ("Aymara", "Aymar aru"), "az": ("Azerbaijani", "Azərbaycan dili"), "ba": ("Bashkir", "Башҡорт теле"), "be": ("Belarusian", "Беларуская мова"), "bg": ("Bulgarian", "Български"), "bh": ("Bihari languages", "भोजपुरी"), "bi": ("Bislama", "Bislama"), "bm": ("Bambara", "Bamanankan"), "bn": ("Bengali", "বাংলা"), "bo": ("Tibetan", "བོད་སྐད་"), "br": ("Breton", "Brezhoneg"), "bs": ("Bosnian", "Bosanski"), "ca": ("Catalan", "Català"), "ce": ("Chechen", "Нохчийн мотт"), "ch": ("Chamorro", "Chamoru"), "co": ("Corsican", "Corsu"), "cr": ("Cree", "ᓀᐦᐃᔭᐍᐏᐣ"), "cs": ("Czech", "Čeština"), "cu": ("Church Slavonic", "Славе́нскїй ѧ҆зы́къ"), "cv": ("Chuvash", "Чӑваш чӗлхи"), "cy": ("Welsh", "Cymraeg"), "da": ("Danish", "Dansk"), "de": ("German", "Deutsch"), "dv": ("Divehi", "ދިވެހިބަސް"), "dz": ("Dzongkha", "རྫོང་ཁ"), "ee": ("Ewe", "Eʋegbe"), "el": ("Greek", "Ελληνικά"), "en": ("English", "English"), "eo": ("Esperanto", "Esperanto"), "es": ("Spanish", "Español"), "et": ("Estonian", "Eesti"), "eu": ("Basque", "Euskara"), "fa": ("Persian", "فارسی"), "ff": ("Fulah", "Fulfulde"), "fi": ("Finnish", "Suomi"), "fj": ("Fijian", "Na Vosa Vakaviti"), "fo": ("Faroese", "Føroyskt"), "fr": ("French", "Français"), "fy": ("Western Frisian", "Frysk"), "ga": ("Irish", "Gaeilge"), "gd": ("Scottish Gaelic", "Gàidhlig"), "gl": ("Galician", "Galego"), "gn": ("Guaraní", "Avañe'ẽ"), "gu": ("Gujarati", "ગુજરાતી"), "gv": ("Manx", "Gaelg"), "ha": ("Hausa", "Hausa"), "he": ("Hebrew", "עברית"), "hi": ("Hindi", "हिन्दी"), "ho": ("Hiri Motu", "Hiri Motu"), "hr": ("Croatian", "Hrvatski"), "ht": ("Haitian Creole", "Kreyòl Ayisyen"), "hu": ("Hungarian", "Magyar"), "hy": ("Armenian", "Հայերեն"), "hz": ("Herero", "Otjiherero"), "ia": ("Interlingua", "Interlingua"), "id": ("Indonesian", "Bahasa Indonesia"), "ie": ("Interlingue", "Interlingue"), "ig": ("Igbo", "Asụsụ Igbo"), "ii": ("Sichuan Yi", "ꆈꌠ꒿ Nuosuhxop"), "ik": ("Inupiaq", "Iñupiaq"), "io": ("Ido", "Ido"), "is": ("Icelandic", "Íslenska"), "it": ("Italian", "Italiano"), "iu": ("Inuktitut", "ᐃᓄᒃᑎᑐᑦ"), "ja": ("Japanese", "日本語"), "jv": ("Javanese", "Basa Jawa"), "ka": ("Georgian", "ქართული"), "kg": ("Kongo", "Kikongo"), "ki": ("Kikuyu", "Gĩkũyũ"), "kj": ("Kuanyama", "Kuanyama"), "kk": ("Kazakh", "Қазақ тілі"), "kl": ("Kalaallisut", "Kalaallisut"), "km": ("Khmer", "ខ្មែរ"), "kn": ("Kannada", "ಕನ್ನಡ"), "ko": ("Korean", "한국어"), "kr": ("Kanuri", "Kanuri"), "ks": ("Kashmiri", "कश्मीरी"), "ku": ("Kurdish", "Kurdî"), "kv": ("Komi", "Коми кыв"), "kw": ("Cornish", "Kernewek"), "ky": ("Kyrgyz", "Кыргызча"), "la": ("Latin", "Latina"), "lb": ("Luxembourgish", "Lëtzebuergesch"), "lg": ("Ganda", "Luganda"), "li": ("Limburgish", "Limburgs"), "ln": ("Lingala", "Lingála"), "lo": ("Lao", "ລາວ"), "lt": ("Lithuanian", "Lietuvių"), "lu": ("Luba-Katanga", "Tshiluba"), "lv": ("Latvian", "Latviešu"), "mg": ("Malagasy", "Malagasy"), "mh": ("Marshallese", "Kajin M̧ajeļ"), "mi": ("Māori", "Te Reo Māori"), "mk": ("Macedonian", "Македонски"), "ml": ("Malayalam", "മലയാളം"), "mn": ("Mongolian", "Монгол"), "mr": ("Marathi", "मराठी"), "ms": ("Malay", "Bahasa Melayu"), "mt": ("Maltese", "Malti"), "my": ("Burmese", "မြန်မာဘာသာ"), "na": ("Nauru", "Dorerin Naoero"), "nb": ("Norwegian Bokmål", "Norsk Bokmål"), "nd": ("North Ndebele", "isiNdebele"), "ne": ("Nepali", "नेपाली"), "ng": ("Ndonga", "Owambo"), "nl": ("Dutch", "Nederlands"), "nn": ("Norwegian Nynorsk", "Norsk Nynorsk"), "no": ("Norwegian", "Norsk"), "nr": ("South Ndebele", "isiNdebele"), "nv": ("Navajo", "Diné bizaad"), "ny": ("Chichewa", "ChiCheŵa"), "oc": ("Occitan", "Occitan"), "oj": ("Ojibwa", "ᐊᓂᔑᓈᐯᒧᐎᓐ"), "om": ("Oromo", "Afaan Oromoo"), "or": ("Odia", "ଓଡ଼ିଆ"), "os": ("Ossetian", "Ирон ӕвзаг"), "pa": ("Punjabi", "ਪੰਜਾਬੀ"), "pi": ("Pāli", "पाळि"), "pl": ("Polish", "Polski"), "ps": ("Pashto", "پښتو"), "pt": ("Portuguese", "Português"), "qu": ("Quechua", "Runa Simi"), "rm": ("Romansh", "Rumantsch"), "rn": ("Kirundi", "Ikirundi"), "ro": ("Romanian", "Română"), "ru": ("Russian", "Русский"), "rw": ("Kinyarwanda", "Ikinyarwanda"), "sa": ("Sanskrit", "संस्कृतम्"), "sc": ("Sardinian", "Sardu"), "sd": ("Sindhi", "سنڌي"), "se": ("Northern Sami", "Davvisámegiella"), "sg": ("Sango", "Yângâ tî sängö"), "si": ("Sinhala", "සිංහල"), "sk": ("Slovak", "Slovenčina"), "sl": ("Slovenian", "Slovenščina"), "sm": ("Samoan", "Gagana Samoa"), "sn": ("Shona", "ChiShona"), "so": ("Somali", "Soomaali"), "sq": ("Albanian", "Shqip"), "sr": ("Serbian", "Српски"), "ss": ("Swati", "SiSwati"), "st": ("Southern Sotho", "Sesotho"), "su": ("Sundanese", "Basa Sunda"), "sv": ("Swedish", "Svenska"), "sw": ("Swahili", "Kiswahili"), "ta": ("Tamil", "தமிழ்"), "te": ("Telugu", "తెలుగు"), "tg": ("Tajik", "Тоҷикӣ"), "th": ("Thai", "ไทย"), "ti": ("Tigrinya", "ትግርኛ"), "tk": ("Turkmen", "Türkmençe"), "tl": ("Tagalog", "Tagalog"), "tn": ("Tswana", "Setswana"), "to": ("Tonga", "Lea fakatonga"), "tr": ("Turkish", "Türkçe"), "ts": ("Tsonga", "Xitsonga"), "tt": ("Tatar", "Татарча"), "tw": ("Twi", "Twi"), "ty": ("Tahitian", "Reo Tahiti"), "ug": ("Uighur", "ئۇيغۇرچە"), "uk": ("Ukrainian", "Українська"), "ur": ("Urdu", "اردو"), "uz": ("Uzbek", "Oʻzbekcha"), "ve": ("Venda", "Tshivenḓa"), "vi": ("Vietnamese", "Tiếng Việt"), "vo": ("Volapük", "Volapük"), "wa": ("Walloon", "Walon"), "wo": ("Wolof", "Wolof"), "xh": ("Xhosa", "isiXhosa"), "yi": ("Yiddish", "ייִדיש"), "yo": ("Yoruba", "Èdè Yorùbá"), "za": ("Zhuang", "Vahcuengh"), "zh": ("Chinese", "中文"), "zu": ("Zulu", "isiZulu")}

def set_locale(language=None):
	target = language or OS_LANGUAGE or ""
	if not target:
		return False

	candidates = [target]
	if "_" in target:
		candidates.append(target.split(".")[0])
		candidates.append(target.split("_")[0])
	if "-" in target:
		candidates.append(target.split("-")[0])
	if target.split("_")[0].split("-")[0].lower() == "en":
		# Ensure a stable dot decimal separator when only a generic English locale is configured.
		candidates.append("C")

	for candidate in candidates:
		try:
			locale.setlocale(locale.LC_ALL, candidate)
			return True
		except locale.Error:
			continue

	try:
		locale.setlocale(locale.LC_ALL, "")
		return True
	except locale.Error:
		return False

def loadTranslation(language):
	# Based on msgfmt.py by Martin v. Löwis: https://github.com/python/cpython/blob/main/Tools/i18n/msgfmt.py
	messages = {}
	section = None
	msgctxt = None
	msgid = b''
	msgstr = b''
	fuzzy = 0
	is_plural = False

	if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
		app_path = os.path.dirname(sys.executable)
	else:
		app_path = os.path.dirname(os.path.abspath(__file__))

	filename = os.path.join(app_path, "locale", f"{language}.po")
	if not os.path.exists(filename):
		raise FileNotFoundError(f"{filename} not found")

	with open(filename, 'r', encoding='utf-8') as f:
		lines = f.readlines()

	for line in lines:
		line = line.strip()
		if not line:
			continue

		if line.startswith('#,') and 'fuzzy' in line:
			fuzzy = 1

		if line.startswith('#'):
			if section == 'STR':
				key = b"%b\x04%b" % (msgctxt, msgid) if msgctxt else msgid
				if not fuzzy and msgstr:
					messages[key] = msgstr
			section = msgctxt = None
			fuzzy = 0
			continue

		if line.startswith('msgctxt'):
			section = 'CTXT'
			msgctxt = b''
			line = line[7:].strip()
		elif line.startswith('msgid') and not line.startswith('msgid_plural'):
			if section == 'STR':
				key = b"%b\x04%b" % (msgctxt, msgid) if msgctxt else msgid
				if not fuzzy and msgstr:
					messages[key] = msgstr
			section = 'ID'
			msgid = msgstr = b''
			is_plural = False
			line = line[5:].strip()
		elif line.startswith('msgid_plural'):
			msgid += b'\0'
			is_plural = True
			line = line[12:].strip()
		elif line.startswith('msgstr'):
			section = 'STR'
			if line.startswith('msgstr['):
				if not is_plural:
					raise ValueError("Plural msgstr found without msgid_plural")
				line = line.split(']', 1)[1].strip()
				if msgstr:
					msgstr += b'\0'
			else:
				if is_plural:
					raise ValueError("Non-indexed msgstr found for plural")
				line = line[6:].strip()

		val = ast.literal_eval(line).encode('utf-8')

		if section == 'CTXT':
			msgctxt += val
		elif section == 'ID':
			msgid += val
		elif section == 'STR':
			msgstr += val

	if section == 'STR':
		key = b"%b\x04%b" % (msgctxt, msgid) if msgctxt else msgid
		if not fuzzy and msgstr:
			messages[key] = msgstr

	keys = sorted(messages.keys())
	offsets = []
	ids = bytearray()
	strs = bytearray()
	for k in keys:
		offsets.append((len(ids), len(k), len(strs), len(messages[k])))
		ids += k + b'\0'
		strs += messages[k] + b'\0'

	keystart = 28 + 16 * len(keys)
	valuestart = keystart + len(ids)
	koffsets = []
	voffsets = []
	for o1, l1, o2, l2 in offsets:
		koffsets += [l1, o1 + keystart]
		voffsets += [l2, o2 + valuestart]

	mo = struct.pack("Iiiiiii", 0x950412de, 0, len(keys), 7 * 4, 7 * 4 + len(keys) * 8, 0, 0)
	mo += array.array("i", koffsets + voffsets).tobytes()
	mo += ids + strs

	return gettext.GNUTranslations(fp=io.BytesIO(mo))

def __(msgid, **kwargs):
	msg = lang.gettext(msgid)
	try:
		return msg.format(**kwargs)
	except (KeyError, IndexError):
		return msgid.format(**kwargs)

def ___(singular, plural, n=1, **kwargs):
	msg = lang.ngettext(singular, plural, n)
	try:
		return msg.format(n=n, **kwargs)
	except (KeyError, IndexError):
		fallback = singular if n == 1 else plural
		return fallback.format(n=n, **kwargs)

def c__(context, msgid, **kwargs):
	try:
		msg = lang.pgettext(context, msgid)
	except AttributeError:
		msg = lang.gettext(msgid)
	try:
		return msg.format(**kwargs)
	except (KeyError, IndexError):
		return msgid.format(**kwargs)

def c___(context, singular, plural, n=1, **kwargs):
	try:
		msg = lang.npgettext(context, singular, plural, n)
	except AttributeError:
		msg = lang.ngettext(singular, plural, n)
	try:
		return msg.format(n=n, **kwargs)
	except (KeyError, IndexError):
		fallback = singular if n == 1 else plural
		return fallback.format(n=n, **kwargs)

def format_number(n):
	return locale.format_string("%d", n, grouping=False)

def format_decimal(value, precision=2, grouping=False, localized=True):
	if localized:
		return locale.format_string("%.{}f".format(precision), value, grouping=grouping)
	else:
		return "{:.{}f}".format(value, precision)

def loadQtTranslation(app=None, language=None):
	global _qt_translator
	try:
		from .pyside import QtCore
	except Exception:
		return False

	if language is None:
		language = CONFIGURED_LANGUAGE or OS_LANGUAGE or ""
	if not language:
		return False

	try:
		qt_locale = QtCore.QLocale(language)
	except Exception:
		qt_locale = QtCore.QLocale.system()
	try:
		QtCore.QLocale.setDefault(qt_locale)
	except Exception:
		pass
	set_locale(qt_locale.name())

	try:
		translations_path = QtCore.QLibraryInfo.path(QtCore.QLibraryInfo.TranslationsPath)
	except AttributeError:
		translations_path = QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.TranslationsPath)

	translator = QtCore.QTranslator(app)
	for catalog in ("qtbase", "qt"):
		if translator.load(qt_locale, catalog, "_", translations_path):
			if app is not None:
				app.installTranslator(translator)
			_qt_translator = translator
			return True

	return False

def init_language(config_path, override=None):
	global lang, CONFIGURED_LANGUAGE, LANGUAGES, TRANSLATION_AUTHOR

	if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
		app_path = os.path.dirname(sys.executable)
	else:
		app_path = os.path.dirname(os.path.abspath(__file__))
	available_langs = ["en"] # Always include English
	available_langs += [os.path.splitext(os.path.basename(f))[0] for f in glob.glob(os.path.join(app_path, "locale", "*.po"))]
	filtered_langs = {code: name for code, name in LANGUAGES.items() if code in available_langs}
	LANGUAGES = dict(sorted(filtered_langs.items(), key=lambda x: x[1][1]))

	try:
		from .IniSettings import IniSettings
	except ImportError:
		from IniSettings import IniSettings

	settings = IniSettings(path=os.path.join(config_path, "settings.ini"))
	if override:
		language_setting = override
		settings.SetValue("Language", override)
	else:
		language_setting = settings.GetValue("Language", default="auto")

	system_language = OS_LANGUAGE.split("_")[0].split("-")[0].lower()

	if language_setting and language_setting.lower() != "auto":
		CONFIGURED_LANGUAGE = language_setting
		lang_code = language_setting.split("_")[0].split("-")[0].lower()
	else:
		CONFIGURED_LANGUAGE = system_language
		if system_language and system_language != "C":
			lang_code = system_language
		else:
			lang_code = "en"

	if lang_code not in LANGUAGES.keys():
		CONFIGURED_LANGUAGE = lang_code = "en"

	if lang_code and lang_code != "en":
		try:
			lang = loadTranslation(lang_code)
			empty = ""
			metadata = lang.gettext(empty)
			for line in metadata.splitlines():
				if line.lower().startswith("last-translator:"):
					TRANSLATION_AUTHOR = line.split(":", 1)[1].strip()
					break
		except Exception as e:
			lang = gettext.NullTranslations()
			print(f"Could not load translation for '{lang_code}': {e}")
	else:
		lang = gettext.NullTranslations()

	set_locale(CONFIGURED_LANGUAGE)

####

try:
	locale.setlocale(locale.LC_ALL, "")
except locale.Error:
	pass

lang_country = None
try:
	lang_country, _ = locale.getlocale()
except (TypeError, ValueError):
	lang_country = None

OS_LANGUAGE = lang_country

if not OS_LANGUAGE and platform.system() == "Darwin":
	try:
		defaults_output = subprocess.check_output(
			["defaults", "read", "-g", "AppleLanguages"],
			stderr=subprocess.DEVNULL,
			text=True,
		).strip()
		apple_languages = []
		for raw_line in defaults_output.splitlines():
			cleaned = raw_line.strip().strip(",").strip().strip('"')
			if cleaned and cleaned not in ("(", ")"):
				apple_languages.append(cleaned)
		if apple_languages:
			OS_LANGUAGE = apple_languages[0]
	except Exception:
		pass

if not OS_LANGUAGE:
	for env_name in ("LC_ALL", "LC_MESSAGES", "LANG"):
		env_value = os.environ.get(env_name)
		if env_value:
			candidate = re.split(r"[.@]", env_value, maxsplit=1)[0].strip()
			if candidate:
				OS_LANGUAGE = candidate
				break

if not OS_LANGUAGE:
	OS_LANGUAGE = "en"

_qt_translator = None
lang = gettext.NullTranslations()



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)
#
# Terminal output, debug logging, and Python exception hook.

import sys, os, re, time, datetime, traceback, platform
from .app import AppInfo, AppContext
from . import i18n
from .i18n import __

class ANSI:
	BOLD = '\033[1m'
	RED = '\033[91m'
	GREEN = '\033[92m'
	YELLOW = '\033[33m'
	DARK_GRAY = '\033[90m'
	RESET = '\033[0m'
	CLEAR_LINE = '\033[2K'

class Logger:
	LOG_ERROR = False

	def __init__(self):
		AppContext.PRINT_LOG.append(
			"FlashGBX {version}\n© 2020–{year} Lesserkuma".format(
				version=AppInfo.VERSION, year=time.strftime("%Y"),
			)
		)

	def write(self, *args, **kwargs):
		msg = "{:s}".format(" ".join(map(str, args)), **kwargs)
		if len(msg.strip()) > 0:
			if ANSI.RED in msg:
				self.LOG_ERROR = True
			AppContext.PRINT_LOG.append(re.sub(r'\x1B\[[0-?]*[ -/]*[@-~]', '', msg.strip()))
			if len(AppContext.PRINT_LOG) > 16 * 1024: AppContext.PRINT_LOG.pop(0)
		sys.__stdout__.write(msg)

	def flush(self):
		sys.__stdout__.flush()

	@classmethod
	def dprint(cls, *args, **kwargs):
		stack = traceback.extract_stack()
		stack = stack[len(stack) - 2]
		msg = "[{:s}] [{:s}:{:d}] {:s}(): {:s}".format(
			str(datetime.datetime.now().astimezone()),
			os.path.split(stack.filename)[1], stack.lineno, stack.name,
			" ".join(map(str, args)), **kwargs,
		)
		AppContext.DEBUG_LOG.append(msg)
		if len(AppContext.DEBUG_LOG) > 64 * 1024: AppContext.DEBUG_LOG.pop(0)
		if AppContext.DEBUG:
			msg = ANSI.CLEAR_LINE + msg
			if isinstance(sys.stdout, Logger):
				temp = sys.stdout
				sys.stdout = sys.__stdout__
				print(msg)
				sys.stdout = temp

	@classmethod
	def write_debug_log(cls, device=False):
		cls.dprint("Now writing debug log file")
		msg = "\n\n\n---- Debug Log ----\n"
		msg += "{:s} version: {:s} ({:d})\n".format(
			AppInfo.NAME, AppInfo.VERSION_PEP440, AppInfo.VERSION_TIMESTAMP,
		)
		msg += "Language: {:s}\n".format(i18n.CONFIGURED_LANGUAGE)
		msg += "Platform: {:s}\n".format(
			AppInfo.os_string() + ", " + platform.machine() + ", " + i18n.OS_LANGUAGE
		)
		if device is not False:
			if device is not None:
				msg += "Connected device: {:s}\n".format(device)
			else:
				msg += "No device connected\n"

		launch_time = datetime.datetime.fromtimestamp(AppContext.LAUNCH_TIMESTAMP).astimezone().replace(microsecond=0)
		now = datetime.datetime.now().astimezone().replace(microsecond=0)
		runtime = now - launch_time
		days, hours, minutes, seconds = (
			runtime.days, runtime.seconds // 3600,
			(runtime.seconds % 3600) // 60, runtime.seconds % 60,
		)
		msg += "Launched: {:s}\n".format(launch_time.isoformat())
		msg += "Log generated: {:s}\n".format(now.isoformat())
		msg += "Runtime: {}d {}h {}m {}s\n\n".format(days, hours, minutes, seconds)

		try:
			fn = AppContext.CONFIG_PATH + os.sep + "debug.log"
			with open(fn, "wb") as f:
				f.write("".encode("UTF-8-SIG"))
				if platform.system() == "Windows":
					f.write("\r\n".join(AppContext.PRINT_LOG).encode("UTF-8"))
					f.write(msg.replace("\n", "\r\n").encode("UTF-8"))
					f.write("\r\n".join(AppContext.DEBUG_LOG).encode("UTF-8"))
				else:
					f.write("\n".join(AppContext.PRINT_LOG).encode("UTF-8"))
					f.write(msg.encode("UTF-8"))
					f.write("\n".join(AppContext.DEBUG_LOG).encode("UTF-8"))
				print(__("The debug log was written to {logfile}", logfile=fn))
			return True
		except:
			return False

	@classmethod
	def exception_hook(cls, exc_type, exc_value, exc_traceback):
		if issubclass(exc_type, KeyboardInterrupt):
			sys.__excepthook__(exc_type, exc_value, exc_traceback)
			return
		s = "EXCEPTION OCCURED\n"
		lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
		for line in lines:
			s += f"{line:s}"
		print(s)
		cls.write_debug_log()
		if isinstance(sys.stdout, Logger):
			sys.stdout.LOG_ERROR = True

# Top-level alias kept because dprint is the most-used helper in the codebase.
dprint = Logger.dprint



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

# pylint: disable=wildcard-import, unused-wildcard-import
import datetime, os, struct, time, zipfile
from .app import AppInfo
from .LK_Device import *
from .i18n import __, c__
from .IniSettings import IniSettings

class GbxDevice(LK_Device):
	DEVICE_NAME = "GBFlash"
	DEVICE_MIN_FW = 1
	DEVICE_MAX_FW = 12
	DEVICE_LATEST_FW_TS = { 5:1780508702, 10:1780508702, 11:1780508702, 12:1780508702, 13:1780508702 }
	PCB_VERSIONS = { 5:'', 12:'v1.2', 13:'v1.3' }
	DEVICE_LABEL_LONG = "GBFlash"
	DEVICE_LABEL_SHORT = "GBFlash"
	FWUPDATE_ACTION = "fwupdate-gbflash"
	CLI_UPDATER_METHOD = "UpdateFirmwareGBFlash"
	DEVICE_SUPPORT_MESSAGE = "For help with your GBFlash, please visit the GitHub page:\nhttps://github.com/simonkwng/GBFlash"

	def __init__(self):
		pass

	def Initialize(self, flashcarts, port=None, max_baud=2000000):
		if self.IsConnected(): self.DEVICE.close()
		conn_msg = []
		ports = []
		if port is not None:
			ports = [ port ]
		else:
			comports = serial.tools.list_ports.comports()
			for i in range(0, len(comports)):
				if comports[i].vid == 0x1A86 and comports[i].pid == 0x7523:
					ports.append(comports[i].device)
			if len(ports) == 0: return False

		for i in range(0, len(ports)):
			if self.TryConnect(ports[i], max_baud):
				self.BAUDRATE = max_baud
				dev = serial.Serial(ports[i], self.BAUDRATE, timeout=0.1)
				self.DEVICE = dev
			else:
				continue

			if self.FW is None or self.FW == {}: continue

			dprint(f"Found a {self.DEVICE_NAME}")
			dprint("Firmware information:", self.FW)
			# dprint("Baud rate:", self.BAUDRATE)

			if self.DEVICE is None or not self.IsConnected():
				self.DEVICE = None
				if self.FW is not None:
					conn_msg.append([0, __("Couldn’t communicate with the {device_name} on port {port}. Please disconnect and reconnect the device, then try again.", device_name=self.DEVICE_NAME, port=ports[i])])
				continue
			elif self.FW is None or self.FW["pcb_ver"] not in self.PCB_VERSIONS.keys() or "cfw_id" not in self.FW or self.FW["cfw_id"] != 'L' or self.FW["fw_ver"] < self.DEVICE_MIN_FW: # Not a CFW by Lesserkuma
				dprint("Incompatible firmware:", self.FW)
				dev.close()
				self.DEVICE = None
				continue
			elif self.FW["fw_ts"] > self.DEVICE_LATEST_FW_TS[self.FW["pcb_ver"]]:
				conn_msg.append([1, __("Note: The {device_name} on port {port} is running a firmware version that is newer than what this version of FlashGBX was developed to work with, so errors may occur.", device_name=self.DEVICE_NAME, port=ports[i])])

			self.MAX_BUFFER_READ = 0x1000
			self.MAX_BUFFER_WRITE = 0x800

			self.PORT = ports[i]
			self.DEVICE.timeout = self.DEVICE_TIMEOUT

			# Load Flash Cartridge Handlers
			self.UpdateFlashCarts(flashcarts)

			# Stop after first found device
			break

		return conn_msg

	def LoadFirmwareVersion(self):
		dprint("Querying firmware version")
		try:
			self.DEVICE.timeout = 0.075
			self.DEVICE.reset_input_buffer()
			self.DEVICE.reset_output_buffer()
			self._write(self.DEVICE_CMD["QUERY_FW_INFO"])
			size = self.DEVICE.read(1)
			self.DEVICE.timeout = self.DEVICE_TIMEOUT
			if len(size) == 0:
				dprint("No response")
				self.FW = None
				return False
			size = struct.unpack("B", size)[0]
			if size != 8: return False
			data = self._read(size)
			info = data[:8]
			keys = ["cfw_id", "fw_ver", "pcb_ver", "fw_ts"]
			values = struct.unpack(">cHBI", bytearray(info))
			self.FW = dict(zip(keys, values))
			self.FW["cfw_id"] = self.FW["cfw_id"].decode('ascii')
			self.FW["fw_dt"] = datetime.datetime.fromtimestamp(self.FW["fw_ts"]).astimezone().replace(microsecond=0).isoformat()
			self.FW["ofw_ver"] = None
			self.FW["pcb_name"] = None
			self.FW["cart_power_ctrl"] = False
			self.FW["bootloader_reset"] = False
			if self.FW["cfw_id"] == "L" and self.FW["fw_ver"] >= 12:
				size = self._read(1)
				name = self._read(size)
				if len(name) > 0:
					try:
						self.FW["pcb_name"] = name.decode("UTF-8").replace("\x00", "").strip()
					except:
						self.FW["pcb_name"] = "Unnamed Device"
					self.DEVICE_NAME = self.FW["pcb_name"]

				# Cartridge Power Control support, Switch Power support, and Switch Mode support
				temp = self._read(1)
				self.FW["cart_power_ctrl"] = True if temp & 1 == 1 else False
				self.FW["cart_presence_switch"] = True if (temp >> 1) & 1 == 1 else False
				self.FW["cart_mode_switch"] = True if (temp >> 2) & 1 == 1 else False

				# Reset to bootloader support
				temp = self._read(1)
				self.FW["bootloader_reset"] = True if temp & 1 == 1 else False
				self.FW["unregistered"] = True if temp >> 7 == 1 else False

			return True

		except Exception as e:
			dprint("Disconnecting due to an error", e, sep="\n")
			try:
				if self.DEVICE.isOpen():
					self.DEVICE.reset_input_buffer()
					self.DEVICE.reset_output_buffer()
					self.DEVICE.close()
				self.DEVICE = None
			except:
				pass
			return False

	def ChangeBaudRate(self, _):
		dprint("Baudrate change is not supported.")

	def GetFirmwareVersion(self, more=False):
		s = "{:s}{:d}".format(self.FW["cfw_id"], self.FW["fw_ver"])
		if self.FW["pcb_name"] is None:
			s = s + " <" + __("unverified") + ">"
		if more:
			return "{base} ({timestamp})".format(base=s, timestamp=self.FW["fw_dt"])
		return s

	def GetFullNameExtended(self, more=False):
		if more:
			return __("{device_name} – Firmware {fw_version} ({timestamp}) on {port}", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), timestamp=self.FW["fw_dt"], port=self.GetPort())
		else:
			return __("{device_name} – Firmware {fw_version} ({port})", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), port=self.GetPort())

	def CanSetVoltageBySwitch(self):
		return False

	def CanSetVoltageByCode(self):
		return True

	def CanSetVoltageByAutoswitch(self):
		return False

	def CanPowerCycleCart(self):
		return self.FW["cart_power_ctrl"]

	def GetSupprtedModes(self):
		return ["DMG", "AGB"]

	def IsSupported3dMemory(self):
		return True

	def IsClkConnected(self):
		return True

	def SupportsFirmwareUpdates(self):
		return True

	def FirmwareUpdateAvailable(self):
		if self.FW["pcb_ver"] == 5 or self.FW["fw_ts"] < 1730592000: # unofficial firmware
			self.FW_UPDATE_REQ = True
			return True
		if self.FW["fw_ts"] != self.DEVICE_LATEST_FW_TS[self.FW["pcb_ver"]]:
			return True
		self.FW_UPDATE_REQ = False
		return False

	def GetFirmwareUpdaterClass(self):
		try:
			return (None, FirmwareUpdaterWindow)
		except:
			return None

	def ResetLEDs(self):
		pass

	def SupportsBootloaderReset(self):
		return self.FW["bootloader_reset"]

	def BootloaderReset(self):
		if not self.SupportsBootloaderReset(): return False
		dprint("Resetting to bootloader...")
		try:
			self._write(self.DEVICE_CMD["BOOTLOADER_RESET"], wait=True)
			self._write(1)
			self.DEVICE.close()
			return True
		except Exception as e:
			print(__("Disconnecting..."), e)
			return False

	def SupportsAudioAsWe(self):
		return not (self.FW["pcb_ver"] < 13 and self.CanPowerCycleCart())

	def GetMode(self):
		if self.FW["fw_ts"] == 1681900614: return self.MODE
		return super().GetMode()

	def SetAutoPowerOff(self, value):
		value &= 0xFFFFFFFF
		return super().SetAutoPowerOff(value)

	def GetFullName(self):
		if self.FW["pcb_ver"] < 13 and self.CanPowerCycleCart():
			s = "{device_name} {pcb_version} + PLUGIN 01".format(device_name=self.GetName(), pcb_version=self.GetPCBVersion())
		else:
			s = "{:s} {:s}".format(self.GetName(), self.GetPCBVersion())
		if self.IsUnregistered():
			s += " (" + __("unregistered") + ")"
		return s

	def GetRegisterInformation(self):
		text = __("Your GBFlash device reported a registration error, which means it may be an illegitimate clone.") + "<br><br>" + __("The device’s integrated piracy detection may limit the device in performance and functionality until proper registration. The FlashGBX software has no control over this.")
		return text


class FirmwareUpdater():
	PORT = None
	DEVICE = None

	def __init__(self, app_path=".", port=None):
		self.APP_PATH = app_path
		self.PORT = port

	def PackPacket(self, packet):
		values = list(packet.values())[:-2]
		data = struct.pack(">IBHHH", *values)
		if packet["payload_len"] > 0:
			data += list(packet.values())[-2]
		data += struct.pack(">I", packet["outro"])
		if len(data) % 2 == 1: data += b'\00'
		return data

	def GetPacket(self):
		hp = 100
		while self.DEVICE.in_waiting == 0:
			time.sleep(0.001)
			hp -= 1
			if hp <= 0:
				return None
		keys = ["intro", "sender", "seq_no", "command", "payload_len"]
		try:
			temp = self.DEVICE.read(11)
			if temp is False: temp = b''
			values = struct.unpack(">IBHHH", temp)
		except (struct.error, TypeError):
			return {"clone":True, "error": "Bootloader error! " + ''.join(format(x, '02X') for x in temp)}
		try:
			data = dict(zip(keys, values))
			data["payload"] = self.DEVICE.read(data["payload_len"])
			data["outro"] = self.DEVICE.read(4)
			if data["outro"] is False: data["outro"] = b''
			data["outro"] = struct.unpack(">I", data["outro"])[0]
		except struct.error:
			return {"clone":True, "error": "Erroneous outro response! " + ''.join(format(x, '02X') for x in data["outro"])}
		return data

	def CRC16(self, data):
		CRCTableAbs = [
			0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
			0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400,
		]
		wCRC = 0xFFFF

		for i in range(len(data)):
			chChar = data[i]
			wCRC = (CRCTableAbs[(chChar ^ wCRC) & 0x0F] ^ (wCRC >> 4))
			wCRC = (CRCTableAbs[((chChar >> 4) ^ wCRC) & 0x0F] ^ (wCRC >> 4))

		return wCRC

	def TryConnect(self, port):
		seq_no = 1
		packet = {
			"intro":0x48484A4A,
			"sender":0,
			"seq_no":seq_no,
			"command":0x21,
			"payload_len":0,
			"payload":bytearray(),
			"outro":0x4A4A4848,
		}
		data = self.PackPacket(packet)

		self.DEVICE = None
		try:
			self.DEVICE = serial.Serial(port, 2000000, timeout=0.5)
			self.DEVICE.write(b'\xF1')
			self.DEVICE.read(1)
			self.DEVICE.write(b'\x01')
			self.DEVICE.close()
			time.sleep(3)
			self.DEVICE = serial.Serial(port, 2000000, timeout=0.5)

		except serial.serialutil.SerialException:
			return False

		self.DEVICE.write(data)
		time.sleep(0.1)
		self.DEVICE.read(self.DEVICE.in_waiting)
		self.DEVICE.write(data)
		data = self.GetPacket()
		if data is None:
			self.DEVICE = None
			return False
		if "error" in data:
			self.DEVICE = None
			return data
		if data["seq_no"] != seq_no:
			self.DEVICE = None
			return False
		if data["command"] != 0x21:
			self.DEVICE = None
			return False
		if struct.unpack(">H", data["payload"][1:3])[0] != 0x03:
			self.DEVICE = None
			return False
		return data

	def WriteFirmware(self, zipfn, fncSetStatus):
		try:
			with zipfile.ZipFile(zipfn) as archive:
				with archive.open("fw.bin") as f: fw_data = bytearray(f.read())
		except (zipfile.BadZipFile, KeyError):
			fncSetStatus(__("The firmware update file is corrupted."))
			return 2

		fncSetStatus(__("Connecting..."))
		data = False
		if self.PORT is None:
			ports = []
			comports = serial.tools.list_ports.comports()
			for i in range(0, len(comports)):
				if comports[i].vid == 0x1A86 and comports[i].pid == 0x7523:
					ports.append(comports[i].device)
			if len(ports) == 0:
				fncSetStatus(__("No device found."))
				return 2

			for port in ports:
				data = self.TryConnect(port)
				if data is not False:
					break
		else:
			data = self.TryConnect(self.PORT)

		if isinstance(data, dict) and "error" in data:
			fncSetStatus(text=data["error"], cloneError="clone" in data and data["clone"] is True)
			return 2
		if not isinstance(data, dict) or self.DEVICE is None:
			fncSetStatus(__("No device found."))
			return 2

		data["program_size"] = struct.unpack(">H", data["payload"][3:5])[0]
		data["page_size"] = struct.unpack(">H", data["payload"][7:9])[0]
		page_size = data["page_size"]
		num_packets = len(fw_data) / page_size
		num_packets = int(-(-num_packets // 1)) # round up

		fncSetStatus(__("Updating firmware..."), setProgress=0)
		seq_no = 2

		pos = 0
		packet_index = 1
		while pos < num_packets:
			buffer = fw_data[pos*page_size:pos*page_size+page_size]
			packet_len = len(buffer)
			buffer += struct.pack(">H", self.CRC16(buffer))
			buffer = struct.pack(">H", packet_len) + buffer
			buffer = struct.pack(">H", packet_index) + buffer

			packet = {
				"intro":0x48484A4A,
				"sender":0,
				"seq_no":seq_no,
				"command":0x24,
				"payload_len":len(buffer),
				"payload":buffer,
				"outro":0x4A4A4848,
			}
			data = self.PackPacket(packet)

			self.DEVICE.write(data)
			data = self.GetPacket()
			if data is None:
				fncSetStatus(__("No response from device."))
				return 2
			if data["seq_no"] != seq_no:
				fncSetStatus(__("Incorrect sequence number."))
				time.sleep(1)
				continue
			if data["command"] != 0x24:
				fncSetStatus(__("Incorrect command."))
				time.sleep(1)
				continue
			if struct.unpack(">H", data["payload"][0:2])[0] != packet_index:
				fncSetStatus(__("Incorrect data packet number."))
				time.sleep(1)
				continue
			if data["payload"][2] != 0x01:
				fncSetStatus(__("Write failed."))
				time.sleep(1)
				continue

			percent = packet_index / num_packets * 100
			fncSetStatus(text=__("Updating firmware... Do not unplug the device!"), setProgress=percent)
			pos += 1
			seq_no += 1
			packet_index += 1

		if pos == num_packets:
			payload = bytearray()
			payload += struct.pack(">H", self.CRC16(fw_data))
			payload += struct.pack(">H", ~self.CRC16(fw_data) & 0xFFFF)
			packet = {
				"intro":0x48484A4A,
				"sender":0,
				"seq_no":seq_no,
				"command":0x23,
				"payload_len":4,
				"payload":payload,
				"outro":0x4A4A4848,
			}
			data = self.PackPacket(packet)
			self.DEVICE.write(data)
			data = self.GetPacket()
			if data is None:
				fncSetStatus(__("No response from device."))
				return 2
			if data["payload"][0] != 1:
				fncSetStatus(text=__("Update failed!"), enableUI=True)

			self.DEVICE.close()
			time.sleep(0.8)
			fncSetStatus(__("Done!"))
			time.sleep(0.2)
			return 1


try:
	from .pyside import QtCore, QtWidgets, QtGui, QDesktopWidget

	class FirmwareUpdaterWindow(QtWidgets.QDialog):
		APP = None
		DEVICE = None
		FWUPD = None
		DEV_NAME = "GBFlash"
		FW_VER = ""
		PCB_VER = ""

		def __init__(self, app, app_path, file=None, icon=None, device=None):
			QtWidgets.QDialog.__init__(self, app)
			if icon is not None: self.setWindowIcon(QtGui.QIcon(icon))
			self.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")
			self.setWindowTitle(AppInfo.NAME + " – " + __("Firmware Updater for {device_name}", device_name="GBFlash"))
			self.setWindowFlags((self.windowFlags() | QtCore.Qt.MSWindowsFixedSizeDialogHint) & ~QtCore.Qt.WindowContextHelpButtonHint)

			self.APP = app
			if device is not None:
				self.FWUPD = FirmwareUpdater(app_path, device.GetPort())
				self.DEV_NAME = device.GetName()
				self.FW_VER = device.GetFirmwareVersion(more=True)
				self.PCB_VER = device.GetPCBVersion()
				self.DEVICE = device
			else:
				self.APP.QT_APP.processEvents()
				self.FWUPD = FirmwareUpdater(app_path, None)

			self.layout = QtWidgets.QGridLayout()
			self.layout.setContentsMargins(-1, 8, -1, 8)
			self.layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
			self.layout_device = QtWidgets.QVBoxLayout()

			# ↓↓↓ Current Device Information
			self.grpDeviceInfo = QtWidgets.QGroupBox(__("Current Firmware"))
			self.grpDeviceInfo.setMinimumWidth(420)
			self.grpDeviceInfoLayout = QtWidgets.QVBoxLayout()
			self.grpDeviceInfoLayout.setContentsMargins(-1, 3, -1, -1)
			rowDeviceInfo1 = QtWidgets.QHBoxLayout()
			self.lblDeviceName = QtWidgets.QLabel(__("Device:"))
			self.lblDeviceName.setMinimumWidth(120)
			self.lblDeviceNameResult = QtWidgets.QLabel("GBFlash")
			rowDeviceInfo1.addWidget(self.lblDeviceName)
			rowDeviceInfo1.addWidget(self.lblDeviceNameResult)
			rowDeviceInfo1.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo1)
			rowDeviceInfo3 = QtWidgets.QHBoxLayout()
			self.lblDeviceFWVer = QtWidgets.QLabel(__("Firmware version:"))
			self.lblDeviceFWVer.setMinimumWidth(120)
			self.lblDeviceFWVerResult = QtWidgets.QLabel("")
			rowDeviceInfo3.addWidget(self.lblDeviceFWVer)
			rowDeviceInfo3.addWidget(self.lblDeviceFWVerResult)
			rowDeviceInfo3.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo3)
			self.grpDeviceInfo.setLayout(self.grpDeviceInfoLayout)
			self.layout_device.addWidget(self.grpDeviceInfo)
			# ↑↑↑ Current Device Information

			# ↓↓↓ Available Firmware Updates
			self.grpAvailableFwUpdates = QtWidgets.QGroupBox(__("Available Firmware"))
			self.grpAvailableFwUpdates.setMinimumWidth(400)
			self.grpAvailableFwUpdatesLayout = QtWidgets.QVBoxLayout()
			self.grpAvailableFwUpdatesLayout.setContentsMargins(-1, 3, -1, -1)

			rowDeviceInfo4 = QtWidgets.QHBoxLayout()
			self.lblDeviceFWVer2 = QtWidgets.QLabel(__("Firmware version:"))
			self.lblDeviceFWVer2.setMinimumWidth(120)
			self.lblDeviceFWVer2Result = QtWidgets.QLabel("(" + __("Please choose the PCB version") + ")")
			rowDeviceInfo4.addWidget(self.lblDeviceFWVer2)
			rowDeviceInfo4.addWidget(self.lblDeviceFWVer2Result)
			rowDeviceInfo4.addStretch(1)
			self.grpAvailableFwUpdatesLayout.addLayout(rowDeviceInfo4)

			self.rowUpdate = QtWidgets.QHBoxLayout()
			self.btnUpdate = QtWidgets.QPushButton(__("Install Firmware Update"))
			self.btnUpdate.setMinimumWidth(200)
			self.btnUpdate.setContentsMargins(20, 20, 20, 20)
			self.connect(self.btnUpdate, QtCore.SIGNAL("clicked()"), lambda: [ self.UpdateFirmware() ])
			self.rowUpdate.addStretch()
			self.rowUpdate.addWidget(self.btnUpdate)
			self.rowUpdate.addStretch()

			self.grpAvailableFwUpdatesLayout.addSpacing(3)
			self.grpAvailableFwUpdatesLayout.addItem(self.rowUpdate)
			self.grpAvailableFwUpdates.setLayout(self.grpAvailableFwUpdatesLayout)
			self.layout_device.addWidget(self.grpAvailableFwUpdates)
			# ↑↑↑ Available Firmware Updates

			self.grpStatus = QtWidgets.QGroupBox("")
			self.grpStatusLayout = QtWidgets.QGridLayout()
			self.prgStatus = QtWidgets.QProgressBar()
			self.prgStatus.setMinimum(0)
			self.prgStatus.setMaximum(1000)
			self.prgStatus.setValue(0)
			self.lblStatus = QtWidgets.QLabel(__("Status: Ready."))

			self.grpStatusLayout.addWidget(self.prgStatus, 1, 0)
			self.grpStatusLayout.addWidget(self.lblStatus, 2, 0)

			self.grpStatus.setLayout(self.grpStatusLayout)
			self.layout_device.addWidget(self.grpStatus)

			self.grpFooterLayout = QtWidgets.QHBoxLayout()
			self.btnClose = QtWidgets.QPushButton(c__("Button (& = Keyboard Shortcut)", "&Close"))
			self.connect(self.btnClose, QtCore.SIGNAL("clicked()"), lambda: [ self.reject() ])
			self.grpFooterLayout.addStretch()
			self.grpFooterLayout.addWidget(self.btnClose)
			self.layout_device.addItem(self.grpFooterLayout)

			self.layout.addLayout(self.layout_device, 0, 0)
			self.setLayout(self.layout)

			self.lblDeviceNameResult.setText(self.DEV_NAME + " " + self.PCB_VER)
			self.lblDeviceFWVerResult.setText(self.FW_VER)
			self.SetPCBVersion()

		def SetPCBVersion(self):
			file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_GBFlash.zip")

			with zipfile.ZipFile(file_name) as zip:
				with zip.open("fw.ini") as f: ini_file = f.read()
				ini_file = ini_file.decode(encoding="utf-8")
				self.INI = IniSettings(ini=ini_file, main_section="Firmware")
				self.OFW_VER = self.INI.GetValue("fw_ver")
				self.OFW_BUILDTS = self.INI.GetValue("fw_buildts")
				self.OFW_TEXT = self.INI.GetValue("fw_text")

			self.lblDeviceFWVer2Result.setText("{:s} ({:s})".format(self.OFW_VER, datetime.datetime.fromtimestamp(int(self.OFW_BUILDTS)).astimezone().replace(microsecond=0).isoformat()))

		def run(self):
			try:
				self.layout.update()
				self.layout.activate()
				screenGeometry = QDesktopWidget().screenGeometry(self)
				x = (screenGeometry.width() - self.width()) / 2
				y = (screenGeometry.height() - self.height()) / 2
				self.move(x, y)
				self.show()
			except:
				return

		def hideEvent(self, event):
			if self.DEVICE is None:
				self.APP.ConnectDevice()
			self.APP.activateWindow()

		def reject(self):
			if self.CloseDialog():
				super().reject()

		def CloseDialog(self):
			if self.btnClose.isEnabled() is False:
				text = __("<b>Warning:</b> If you close this window while a firmware update is still running, it might leave the device in an unbootable state.") + " " + __("You can still recover it by running the Firmware Updater again later.") + "<br><br>" + __("Are you sure you want to close this window?")
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Warning, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
				msgbox.setDefaultButton(QtWidgets.QMessageBox.No)
				answer = msgbox.exec()
				if answer == QtWidgets.QMessageBox.No: return False
			return True

		def UpdateFirmware(self):
			file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_GBFlash.zip")

			if self.APP.CONN is None or self.APP.CONN.BootloaderReset() is False:
				self.APP.DisconnectDevice()
				text = __("Please follow these steps to proceed with the firmware update:")
				text += "\n\n" + __(
					"- Unplug your GBFlash device.\n"
					"- On your GBFlash circuit board, push and hold the small button (U22) while plugging the USB cable back in.\n"
					"- If done right, the blue LED labeled “ACT” should now keep blinking twice continuously."
				)
				text += "\n" + __("- Click OK to continue.")
				text += "\n\n" + __("Note: Illegitimate clones of the GBFlash may have been modified to disallow firmware updates.")
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Information, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
				msgbox.setDefaultButton(QtWidgets.QMessageBox.Ok)
				answer = msgbox.exec()
				if answer == QtWidgets.QMessageBox.Cancel: return
			else:
				self.APP.DisconnectDevice()
				time.sleep(1)

			self.btnUpdate.setEnabled(False)
			self.btnClose.setEnabled(False)

			while True:
				ret = self.FWUPD.WriteFirmware(file_name, self.SetStatus)
				if ret == 1:
					text = __("The firmware update is complete!")
					if self.PCB_VER != "v1.3":
						text += "\n\n" + __("Please re-connect the USB cable now.")
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Information, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					self.DEVICE = None
					self.reject()
					return True
				elif ret == 2:
					text = __("The firmware update has failed. Please try again.")
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return False
				elif ret == 3:
					text = __("The firmware update file is corrupted. Please re-install the application.")
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return False

		def SetStatus(self, text, enableUI=False, setProgress=None, cloneError=False):
			self.lblStatus.setText(__("Status: {text}", text=text))

			if cloneError:
				text = __("Your GBFlash device failed to enter the firmware update mode. This means your GBFlash may be an <b>illegitimate clone</b> that blocks certain features intentionally. If this error persists, return the device for a refund.") + "<br><br>" + text
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
				msgbox.exec()
				return False

			if setProgress is not None:
				self.prgStatus.setValue(setProgress * 10)
			if enableUI:
				self.btnUpdate.setEnabled(True)
				self.btnClose.setEnabled(True)
			self.APP.QT_APP.processEvents()
except ImportError:
	pass



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

from .pyside import QtCore, QtWidgets, QtGui, QDesktopWidget
from .i18n import __, c__
from .InteractiveConsole import InteractiveConsole
from .app import AppInfo

class InteractiveConsoleWindow(QtWidgets.QDialog):
	APP = None
	CONN = None
	MODE = None
	IM = None

	def __init__(self, app, icon=None):
		QtWidgets.QDialog.__init__(self, app)
		if icon is not None: self.setWindowIcon(QtGui.QIcon(icon))
		self.setWindowTitle(AppInfo.NAME + " – " + __("Interactive Console"))
		flags = self.windowFlags()
		flags = (flags & ~QtCore.Qt.WindowContextHelpButtonHint) | QtCore.Qt.WindowCloseButtonHint | QtCore.Qt.WindowMaximizeButtonHint
		self.setWindowFlags(flags)
		self.resize(700, 400)

		self.APP = app
		self.CONN = app.CONN
		self.MODE = self.CONN.GetMode()
		self.IM = InteractiveConsole(self.CONN, on_output=self.AppendOutput, on_error=self.AppendOutput)

		self.layout = QtWidgets.QVBoxLayout()
		self.layout.setContentsMargins(8, 8, 8, 8)

		mono_font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
		mono_font.setStyleHint(QtGui.QFont.TypeWriter)

		self.txtOutput = QtWidgets.QPlainTextEdit()
		self.txtOutput.setReadOnly(True)
		self.txtOutput.setFont(mono_font)
		self.txtOutput.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
		self.layout.addWidget(self.txtOutput)

		rowInput = QtWidgets.QHBoxLayout()
		self.lblPrompt = QtWidgets.QLabel(">")
		self.lblPrompt.setFont(mono_font)
		self.txtInput = QtWidgets.QLineEdit()
		self.txtInput.setFont(mono_font)
		self.txtInput.returnPressed.connect(self.OnSubmit)
		rowInput.addWidget(self.lblPrompt)
		rowInput.addWidget(self.txtInput)
		self.layout.addLayout(rowInput)

		rowButtons = QtWidgets.QHBoxLayout()
		rowButtons.addStretch()
		self.btnClose = QtWidgets.QPushButton(c__("Button (& = Keyboard Shortcut)", "&Close"))
		self.btnClose.setStyleSheet("padding: 5px 15px;")
		self.btnClose.setAutoDefault(False)
		self.btnClose.setDefault(False)
		self.btnClose.clicked.connect(self.reject)
		rowButtons.addWidget(self.btnClose)
		self.layout.addLayout(rowButtons)

		self.setLayout(self.layout)

		self.History = []
		self.HistoryIndex = 0
		self.txtInput.installEventFilter(self)

	def run(self):
		self.CONN.SetAutoPowerOff(value=0)
		self.CONN.CartPowerOn()
		self.IM.print_help()
		self.txtInput.setFocus()
		self.layout.update()
		self.layout.activate()
		screenGeometry = QDesktopWidget().screenGeometry(self)
		x = (screenGeometry.width() - self.width()) / 2
		y = (screenGeometry.height() - self.height()) / 2
		self.move(x, y)
		self.show()

	def hideEvent(self, event):
		try:
			self.APP.SetAutoPowerOff()
		except:
			pass
		self.APP.activateWindow()

	def eventFilter(self, obj, event):
		if obj is self.txtInput and event.type() == QtCore.QEvent.KeyPress:
			if event.key() == QtCore.Qt.Key_Up:
				if self.History and self.HistoryIndex > 0:
					self.HistoryIndex -= 1
					self.txtInput.setText(self.History[self.HistoryIndex])
				return True
			elif event.key() == QtCore.Qt.Key_Down:
				if self.History and self.HistoryIndex < len(self.History) - 1:
					self.HistoryIndex += 1
					self.txtInput.setText(self.History[self.HistoryIndex])
				else:
					self.HistoryIndex = len(self.History)
					self.txtInput.clear()
				return True
		return super().eventFilter(obj, event)

	def AppendOutput(self, text):
		self.txtOutput.appendPlainText(text)
		self.txtOutput.verticalScrollBar().setValue(self.txtOutput.verticalScrollBar().maximum())

	def OnSubmit(self):
		line = self.txtInput.text().strip()
		self.txtInput.clear()
		if not line:
			self.AppendOutput("")
			return
		self.AppendOutput("> " + line)
		if not self.History or self.History[-1] != line:
			self.History.append(line)
		self.HistoryIndex = len(self.History)

		self.txtInput.setEnabled(False)
		self.btnClose.setEnabled(False)
		try:
			if not self.IM.execute_line(line):
				self.reject()
				return
		finally:
			self.txtInput.setEnabled(True)
			self.btnClose.setEnabled(True)
			self.txtInput.setFocus()



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

import time, threading, statistics, random

class Progress():
	MUTEX = threading.Lock()
	PROGRESS = {}
	UPDATER = None
	WAITER = None

	def __init__(self, updater, waiter):
		self.UPDATER = updater
		self.WAITER = waiter

	def _is_outlier(self, speeds, new_number, threshold):
		if not speeds: return False
		mean = sum(speeds) / len(speeds)
		std_deviation = (sum((x - mean) ** 2 for x in speeds) / len(speeds)) ** 0.5
		lower_bound = mean - threshold * std_deviation
		upper_bound = mean + threshold * std_deviation
		if new_number < lower_bound or new_number > upper_bound:
			return True
		else:
			return False

	def SetProgress(self, args):
		self.MUTEX.acquire(1)
		try:
			if not "method" in self.PROGRESS: self.PROGRESS = {}
			now = time.time()
			if args["action"] == "USER_ACTION":
				self.WAITER(args)

			elif args["action"] == "INITIALIZE":
				self.PROGRESS["action"] = args["action"]
				self.PROGRESS["method"] = args["method"]
				if "voltage" in args:
					self.PROGRESS["voltage"] = args["voltage"]
				elif "voltage" in self.PROGRESS:
					del self.PROGRESS["voltage"]
				if "flash_offset" in args:
					self.PROGRESS["flash_offset"] = args["flash_offset"]
				else:
					self.PROGRESS["flash_offset"] = 0
				if "size" in args:
					self.PROGRESS["size"] = args["size"] - self.PROGRESS["flash_offset"]
				else:
					self.PROGRESS["size"] = 0
				if "pos" in args:
					self.PROGRESS["pos"] = args["pos"] - self.PROGRESS["flash_offset"]
				else:
					self.PROGRESS["pos"] = 0
				if "sector_count" in args:
					self.PROGRESS["sector_count"] = args["sector_count"]
				else:
					self.PROGRESS["sector_count"] = 1
				if "time_start" in args:
					self.PROGRESS["time_start"] = args["time_start"]
				else:
					self.PROGRESS["time_start"] = now
				if "abortable" in args:
					self.PROGRESS["abortable"] = args["abortable"]
				else:
					self.PROGRESS["abortable"] = True
				self.PROGRESS["time_last_emit"] = now
				self.PROGRESS["time_last_update_speed"] = now
				self.PROGRESS["time_left"] = 0
				self.PROGRESS["speed"] = 0
				self.PROGRESS["speeds"] = []
				self.PROGRESS["bytes_last_update_speed"] = 0
				self.PROGRESS["sector_erase_time"] = 0
				self.UPDATER(self.PROGRESS)

			if args["action"] == "ABORT":
				self.UPDATER(args)
				self.PROGRESS = {}

			elif args["action"] in ("ERASE", "SECTOR_ERASE", "UNLOCK", "UPDATE_RTC", "CALC_CHECKSUMS", "ERROR"):
				if "time_start" in self.PROGRESS:
					args["time_elapsed"] = now - self.PROGRESS["time_start"]
				elif "time_start" in args:
					args["time_elapsed"] = now - args["time_start"]
				args["pos"] = 1
				args["size"] = 0
				self.UPDATER(args)

			elif self.PROGRESS == {}:
				return

			elif args["action"] in ("READ", "WRITE", "UPDATE_POS"):
				if "method" not in self.PROGRESS: return
				elif args["action"] == "READ" and self.PROGRESS["method"] in ("SAVE_WRITE", "ROM_WRITE"): return
				elif args["action"] == "WRITE" and self.PROGRESS["method"] in ("SAVE_READ", "ROM_READ", "ROM_WRITE_VERIFY"): return
				if self.PROGRESS["pos"] > self.PROGRESS["size"]: return
				skip_speed = False
				self.PROGRESS["action"] = "PROGRESS"
				if args["action"] in ("READ", "WRITE"):
					self.PROGRESS["pos"] += args["bytes_added"]
				elif args["action"] == "UPDATE_POS":
					if self.PROGRESS["pos"] == args["pos"] - self.PROGRESS["flash_offset"]:
						skip_speed = True
					if "skipping" in args and args["skipping"] is True:
						skip_speed = True
					self.PROGRESS["pos"] = args["pos"] - self.PROGRESS["flash_offset"]
					if "sector_erase_time" in args:
						if "sector_erase_time" in self.PROGRESS:
							self.PROGRESS["sector_erase_time"] = (self.PROGRESS["sector_erase_time"] + args["sector_erase_time"]) / 2
						else:
							self.PROGRESS["sector_erase_time"] = args["sector_erase_time"]
					if "sector_pos" in args:
						if "sector_erase_time" not in args: self.PROGRESS["sector_erase_time"] = 0
						self.PROGRESS["sector_pos"] = args["sector_pos"]
					if "abortable" in args:
						self.PROGRESS["abortable"] = args["abortable"]

				if ((now - self.PROGRESS["time_last_emit"]) > 0.06) or "force_update" in args and args["force_update"] is True:
					self.PROGRESS["time_elapsed"] = now - self.PROGRESS["time_start"]
					time_delta = now - self.PROGRESS["time_last_update_speed"]
					pos_delta = self.PROGRESS["pos"] - self.PROGRESS["bytes_last_update_speed"]
					if time_delta > 0 and (time.time() - self.PROGRESS["time_start"] > 2) and "sector_erase_time" not in args:
						speed = (pos_delta / time_delta) / 1024
						if speed > 0 and not skip_speed:
							if len(self.PROGRESS["speeds"]) < 40 or not self._is_outlier(speeds=self.PROGRESS["speeds"], new_number=speed, threshold=25):
								self.PROGRESS["speeds"].append(speed)
							if len(self.PROGRESS["speeds"]) > 50: self.PROGRESS["speeds"].pop(0)
							if len(self.PROGRESS["speeds"]) > 1 and random.randint(0, 10) == 0: self.PROGRESS["speeds"].pop(0)
						if len(self.PROGRESS["speeds"]) > 0:
							self.PROGRESS["speed"] = statistics.mean(self.PROGRESS["speeds"])
						else:
							self.PROGRESS["speed"] = 0
					self.PROGRESS["time_last_update_speed"] = now
					self.PROGRESS["bytes_last_update_speed"] = self.PROGRESS["pos"]

					if "skipping" in args and args["skipping"] is True:
						self.PROGRESS["speed"] = 0
						self.PROGRESS["skipping"] = True
					else:
						self.PROGRESS["skipping"] = False

					if self.PROGRESS["speed"] > 0 and len(self.PROGRESS["speeds"]) > 0:
						total_speed = self.PROGRESS["speed"]
						self.PROGRESS["time_left"] = (self.PROGRESS["size"] - self.PROGRESS["pos"]) / 1024 / total_speed
						if self.PROGRESS["sector_erase_time"] > 0:
							self.PROGRESS["time_left"] += self.PROGRESS["sector_erase_time"] * (self.PROGRESS["sector_count"] - self.PROGRESS["sector_pos"])

					self.UPDATER(self.PROGRESS)
					self.PROGRESS["time_last_emit"] = now

			elif args["action"] == "UPDATE_INFO":
				self.PROGRESS["text"] = args["text"]
				self.PROGRESS["action"] = args["action"]
				self.UPDATER(self.PROGRESS)

			elif args["action"] == "FINISHED":
				self.PROGRESS["pos"] = self.PROGRESS["size"]
				self.UPDATER(self.PROGRESS)
				self.PROGRESS["action"] = args["action"]
				self.PROGRESS["bytes_last_update_speed"] = self.PROGRESS["size"]
				self.PROGRESS["time_elapsed"] = now - self.PROGRESS["time_start"]
				self.PROGRESS["time_last_emit"] = now
				self.PROGRESS["time_last_update_speed"] = now
				self.PROGRESS["time_left"] = 0
				if self.PROGRESS["time_elapsed"] == 0: self.PROGRESS["time_elapsed"] = 0.001
				self.PROGRESS["speed"] = (self.PROGRESS["size"] / self.PROGRESS["time_elapsed"]) / 1024
				self.PROGRESS["bytes_last_emit"] = self.PROGRESS["size"]
				if "verified" in args:
					self.PROGRESS["verified"] = (args["verified"] == True)

				if self.PROGRESS["speed"] > self.PROGRESS["size"] / 1024:
					self.PROGRESS["speed"] = self.PROGRESS["size"] / 1024

				self.UPDATER(self.PROGRESS)
				del(self.PROGRESS["method"])

		finally:
			self.MUTEX.release()
